//
// Author : Mark Stephenson
// Purpose: Inserts library calls before and after innermost loops.
// 
// Rather than commenting the header file while leaving the code 
// mostly uncommented, I'm taking an alternate approach; I'm commenting
// the code itself.


#ifndef __LOOP_PROFILE_H
#define __LOOP_PROFILE_H

#include "cg_region.h"
#include "findloops.h"
#include "cxx_memory.h"
#include "profile_com.h"
#include "tn.h"
#include "instr_reader.h"
#include "glob.h"
#include "fb_freq.h"
#include "fb_info.h"
#include "const.h"
#include "op.h"
#include "mempool.h"
#include "profile_util.h"
#include "wn.h"
#include <list>
#include <map>


extern void CG_Loop_Instrument (CGRIN* rin, PROFILE_PHASE phase, BOOL add_checks);
extern void Dot_Print (char*);


class LOOP_INSTRUMENT
{
 public:

  LOOP_INSTRUMENT (CGRIN *rin, PROFILE_PHASE phase, BOOL checks);
  ~LOOP_INSTRUMENT ();

  void Instrument ();

 protected:

  typedef struct 
  {
    BB *exit_node;
    BB *target_node;
    BB *fall_through;
  } EDGE_INFO;

  typedef map<BB*,BB*> BB_ASSOCS;
  typedef list<EDGE_INFO> EDGE_LIST;

 protected:

  bool Skip_PU ();

  void Create_Instrumented_Prolog (LOOP_DESCR*, INT);
  BB*  Create_Prolog (LOOP_DESCR*);
  BB*  Create_Empty_Prolog (LOOP_DESCR*, BB*);

  void Create_Instrumented_Epilog (LOOP_DESCR*, INT);
  void Create_Epilogs (LOOP_DESCR*, BB_ASSOCS&);
  void Create_Epilog_Blocks (BB_ASSOCS&, BB_ASSOCS&, EDGE_LIST&, LOOP_DESCR*);
  void Insert_Epilog_Blocks (BB_ASSOCS&, BB_ASSOCS&, EDGE_LIST&, LOOP_DESCR*);
  BB*  Non_Loop_Fall_Thru_Pred (BB*, LOOP_DESCR*);

  ST*  Get_Global_ST (char*);
  void Create_Global_Symbols ();
  void Start_Timer (BB*, INT);
  void Stop_Timer (BB*, INT);
  void Consistency_Check (BB*, TN*, INT, INT);

  BB* Gen_And_Set_BB_Like (BB*);

  INT  read_gid ();
  void write_gid (INT);
  void write_associations (char*, INT, INT, INT);

  // Debugging procedures.
  void Print_BB (BB*);

 protected:

  static char *LOOP_TIMERS_ARR;
  static char *LOOP_TIMERS_SIZE;
  static char *LOOP_TIMERS_ERROR;

  CGRIN *m_rin;
  PROFILE_PHASE m_phase;
  BOOL m_measure_outermost;
  BOOL m_perform_checks;
  MEM_POOL m_mempool;
  REGISTER m_stack_reg;
  TN *m_time_arr_base_offset;
  TN *m_time_size_offset;
  TN *m_error_offset;
  TN *m_itc_tn;
  char *m_srcfile_pu_name;
};

#endif
